!--------------------------------------------------------------------------------
!	Yun Wang
!--------------------------------------------------------------------------------
!***************************************************************************
!	This code implements flux-averaging for SNe with correlated errors *
!***************************************************************************	
!	It computes the chi^2 = - 2*log(likelihood) of a model	           *
!	with given E(z) using flux-averaged type Ia supernova data.	   *
!	Only the Riess 2004 sample is included at present. It will         *
!	be expanded to include all available SN data in the future.	   *   			
!									   *
!	To use this code:						   *
!	(1) Replace the function invEz(z) with the 1/E(z)                  *
!	from your model, note that 					   *
!									   *
!   	[E(z)]^2 = Omega_m * (1+z)^3 + Omega_k * (1+z)^2 + Omega_X* X(z),  *
!   	where the dark energy function X(z) = rho_X(z) / rho_X(0)	   *
!									   *
!	(2) Marginalize over the SN nuisance parameters 		   *
!	(alpha, beta, Msn) by varying them.   			           *
!	   								   *
!	Report any problems to wang@nhn.ou.edu.	   	                   *
!									   *
!***************************************************************************
!	If you use this code, please reference these three papers:	   *
!	   								   *
!	(1) Yun Wang, "Flux-averaging Analysis of Type Ia Supernova Data", *
!	    ApJ 536, 531 (2000), astro-ph/9907405.			   *
!		   							   *
!	(2) Yun Wang, and Pia Mukherjee, "Model-Independent Constraints on *
!	    Dark Energy Density from Flux-averaging Analysis of Type Ia    *
!	    Supernova Data", ApJ, in press (2004), astro-ph/0312192        *
!						   			   *
!	(3) Yun Wang, Chia-Hsun Chuang, and Pia Mukherjee, 2011,	   *
!	  	   							   * 
!***************************************************************************
!	v1: 7/14/2011					                   *				
!	    The SN sample is compiled by Conley et al. (2011)              *
!***************************************************************************
! Yun Wang, Oct 18, 2011: added clarification on  (alpha, beta, Msn) !                                !
!--------------------------------------------------------------------!

	program snfluxavg2
	implicit none
	integer i, ndata, np, j, s, q
	parameter(np=1500)	
	integer nmin, nbin, count, nq, i1, i2, counta(100), set
	integer SN_num, SNa_nzlow(np), SNa_nzhigh(np), indx(np)
	real siginta(0:3), zhel(np),str(np),ds(np),color(np),dcolor(np)
	real thirdvar,d3rdvar,cov_ms(np),cov_mc(np),cov_sc(np),Dii0(np) 
	real zdata(np), dzdata(np),sum0, z, zmean, Msn, Dii(np)
	real rcov0, rcov0a(np), facz, H0dLd, eps,Gamma,del
	real fluxd(np), mudat(np), dmudat(np)
	real alpha, beta, rcov0bina(np), dzfluxbin, ex 
	real sumL, invEz, pz, zmin, zmax, w0, wa, kappa
	real faczmean, zabin(np), fluxbin(np)
	real xsn(np), ysn(np), drdz, Omegam, Omegak, sigmz2
	real zlowa(np), zhigha(np), dzhighmin(np), dzlowmin(np)
	real dzlow, dzhigh, zlowprev(np), zhighprev(np), H0dLa(100)
	real H0dL, lumd(np), y0a(100), covmatF(np,np), invcovmatF(np,np)
	real sumc, sums, sum, sumac, sumas, sumcs, za(np), chisq
	real*8 v0(np,np), v0a(np,np), v0b(np,np), va(np,np)
	real*8 vb(np,np), vab(np,np), covmat(np,np) 
	real*8 vv2(np,np), nn0(np,np)
	real*8 check(np,np)	
	character*80 fileold, filecovSN(10)
	character*12 char(np)
		
	common /nmin/ nmin
	common /cospar/ Omegam, Omegak, w0, wa	
		    	
	external ex, invEz, intg, matinvert
	      		        		
	open(unit=5,file='snfluxavg2.inp',status='old')

	read(5,*) fileold, ndata !SNLS3
	read(5,*) dzfluxbin, nq
	
	eps=1.e-6
	nmin=10
			  
	do i=1,6
	   read(5,*) filecovSN(i)
	   write(*,*) filecovSN(i)
	end do
	  
	read(5,*) alpha, beta, Msn, Omegam, Omegak, w0, wa
	write(*,*) 'alpha, beta, Msn, Omegam, Omegak, w0, wa='	
	write(*,*) alpha, beta, Msn, Omegam, Omegak, w0, wa

        siginta(0)=0.068
        siginta(1)=0.113
        siginta(2)=0.082
        siginta(3)=0.099
        
        open(unit=2,file=fileold,status='old')
        
	do i=1,ndata    	         
     	    read(2,*)char(i),zdata(i),zhel(i),dzdata(i),mudat(i),dmudat(i),
     &		str(i),ds(i),color(i),dcolor(i),thirdvar,d3rdvar,cov_ms(i),
     &		cov_mc(i),cov_sc(i),set  
     
     	   pz=0.0005 ! 150km/s
     	   dzdata(i)=sqrt(dzdata(i)**2+pz*pz) ! adding peculiar velocity residuls
	   Dii0(i)=dmudat(i)**2+siginta(set)**2    
	   
	   if (zdata(i).lt.zmin) zmin=zdata(i)
	   if (zdata(i).gt.zmax) zmax=zdata(i)	 	         	       
     	end do   
     	close(2)  	

	write(*,*) 'zmin,zmax=',zmin,zmax
	
     	  do s=1,6
     	    open(unit=50+s,file=filecovSN(s),status='old')
     	  end do
     	  
     	  read(51,*) s
     	  do i=1,ndata
     	    read(51,*) (v0(i,j),j=1,ndata)  	      
     	  end do !i-loop
     	  
     	  read(52,*) s
     	  do i=1,ndata
     	    read(52,*) (v0a(i,j),j=1,ndata)	      
     	  end do !i-loop     	  
     	  
      	  read(53,*) s
     	  do i=1,ndata
     	    read(53,*) (v0b(i,j),j=1,ndata)      
     	  end do !i-loop     	      	  

     	  read(54,*) s
     	  do i=1,ndata
     	    read(54,*) (va(i,j),j=1,ndata)	      
     	  end do !i-loop     	  
     	       	         
     	  read(55,*) s
     	  do i=1,ndata
     	    read(55,*) (vb(i,j),j=1,ndata)	      
     	  end do !i-loop     	       	       	         

     	  read(56,*) s
     	  do i=1,ndata
     	    read(56,*) (vab(i,j),j=1,ndata)    	      
     	  end do !i-loop     	  
     	       	         	                	          	         	      
          do s=1,6
     	    close(50+s)
     	  end do !s-loop
     			
	count=0
	sumac=0.
	sumas=0.
	sumc=0.
	sums=0.
	sumcs=0.
 
	kappa= sqrt(abs(Omegak))	
	       	        
        chisq=0.
               
!	converting to flux, tabulate r_cov(j):
	do j=1,ndata
	   z=zdata(j)
	            
	        call intg(0.,z,invEz,eps,Gamma,del)  

	        if (Omegak.lt.0.0) then ! closed universe
	            rcov0 = sin(kappa*Gamma)/kappa
	        else if (Omegak.gt.0.) then ! open universe
	            rcov0 = sinh(kappa*Gamma)/kappa
	        else 
	            rcov0=Gamma	         
	        end if		
	   	     
	   H0dL=rcov0*(1.0+z) !H0*dL	   
	   	       
	   rcov0a(j)=rcov0
	      	           
	        if (Omegak.lt.0.0) then ! closed universe
	            drdz = cos(kappa*Gamma)*invEz(z)
	        else if (Omegak.gt.0.) then ! open universe
	            drdz = cosh(kappa*Gamma)*invEz(z)
	        else 
	      	    drdz=invEz(z) !flat universe      
	        end if	      
	      	      
	      sigmz2=((5./log(10.))*(1./(1.+z)+drdz/rcov0))**2 * 
     &		     dzdata(j)**2      !c=3.e5km/s   
	      

		Dii(j) =Dii0(j)+ (alpha*ds(j))**2+ (beta*dcolor(j))**2 +                               
     &	 		2.*alpha*cov_ms(j)- 2.* beta*cov_mc(j) - 
     &			2.*alpha*beta*cov_sc(j) + sigmz2  
     

	      H0dLd= ex( log(10.)*(mudat(j)-(Msn-
     &			alpha*(str(j)-1.0)+beta*color(j)) )/5.)     
	      
	     
	      fluxd(j)=1./H0dLd**2
	      lumd(j)=fluxd(j)*H0dL**2
	                    	      
	end do

!	sum up snls3 cov mat and invert it:
	   
          do i=1,ndata
            do j=i,ndata            
              covmat(i,j)=v0(i,j)+alpha**2*va(i,j) + beta**2*vb(i,j) + 
     &		2.*alpha*v0a(i,j) - 2.*beta*v0b(i,j) - 
     &		2.*alpha*beta*vab(i,j)
     	      if (i.eq.j) then
     	          covmat(i,j)=covmat(i,j)+Dii(i)
     	      end if
     	      
     	      covmat(j,i)=covmat(i,j)
     	      
            end do  
          end do        


!	binning:

	  za(1)=0.0
	  do i=1,nq
	      za(i+1)=za(1)+ dzfluxbin*real(i)
	      zlowprev(i+1)=za(i+2)
	  end do	  

	  nbin=0
            
	  SN_num=0   
        	  
	  do q=1,nq
	    sumL=0.
	    zmean=0.
	    count=0
	    	    
	    dzlowmin(q)=0.1
	    dzhighmin(q)=0.1     	    
	    
	    do j=1,ndata
		z=zdata(j)

		if (z.gt.za(q).and.z.le.za(q+1)) then
		    
		    count=count+1
		    SN_num=SN_num+1
		    
	            rcov0=rcov0a(j)    
	                        	   		    				   
 		    facz= ((1.+z)*rcov0)**2 
	          		    				   
		    sumL=sumL+ fluxd(j)*facz
		    zmean=zmean+z

		    dzlow=abs(z-za(q)) 	
		    if (dzlow.le.dzlowmin(q)) then
		      zlowa(q)=z	    
		      SNa_nzlow(q)=SN_num
		      dzlowmin(q)=dzlow	
		      zlowprev(q)=z	    
		    end if
		    
		    dzhigh=abs(z-za(q+1)) 	
		    if (dzhigh.le.dzhighmin(q).and.z.ge.zhighprev(q)) then
		      zhigha(q)=z	    
		      SNa_nzhigh(q)=SN_num
		      dzhighmin(q)=dzhigh
		      zhighprev(q)=z		    
		    end if
		    
		end if
	    end do

	    SNa_nzlow(1)=1
	    if (count.gt.0) then
     
    	      if (q.gt.1.and.SNa_nzlow(q).gt.(SNa_nzhigh(q-1)+1).and.
     &			SNa_nzhigh(q-1).gt.0) then 

     	        SNa_nzlow(q)=SNa_nzhigh(q-1)+1

     	      end if
     	      
     	      counta(q)=count

     	    end if

	    if (count.gt.0) then
		nbin=nbin+1
		zmean=zmean/real(count)

		call intg(0.,zmean,invEz,eps,Gamma,del)  
		
	        if (Omegak.lt.0.0) then ! closed universe
	            rcov0 = sin(kappa*Gamma)/kappa
	        else if (Omegak.gt.0.) then ! open universe
	            rcov0 = sinh(kappa*Gamma)/kappa
	        else 
	            rcov0=Gamma	         
	        end if
	        	      
 		faczmean= ((1.+zmean)*rcov0)**2  		   
 		 		    		
	        sum0= sumL/( faczmean*dble(count) )
		        
		rcov0bina(nbin)=rcov0      
	        zabin(nbin)=zmean
	        fluxbin(nbin)=sum0
	    end if

	  end do	!end binning


!-----------------------------------------------------------------	    
!	compute covariance matrix for correlated errors in SNe:	

	do i=1,nbin
	  xsn(i)=zabin(i)
	  	
	  z=xsn(i)
	  rcov0=rcov0bina(i)
	  H0dLa(i)=(1.+z)*rcov0	
	  y0a(i)=1./H0dLa(i)**2 ! predicted "flux"		  

	  ysn(i)=fluxbin(i)		  
	end do
	
	do i=1,nbin
	  do j=1,nbin

	    sum=0.	  	    
	    do i1=SNa_nzlow(i), SNa_nzhigh(i) !1, counta(j)
	      do i2=SNa_nzlow(j), SNa_nzhigh(j)
	      	sum=sum+lumd(i1)*lumd(i2)*covmat(i1,i2)
	      end do
	    end do    
	    
	    covmatF(i,j)=sum*(log(10.)/(2.5*H0dLa(i)*H0dLa(j)))**2 /
     &				(counta(i)*counta(j))
            nn0(i,j)=covmatF(i,j)
	  end do 
	end do   
	
!	do i=1,nbin
!	  write(99,*) (nn0(i,j),j=1,nbin)
!	end do
	
!	invert it:	
	call matinvert(nn0,vv2,indx,nbin,np)
	
	do i=1,nbin
	  do j=1,nbin
	    invcovmatF(i,j)=nn0(i,j)
	  end do
	end do
	
	do i=1,nbin
	  do j=1,nbin
	    sum=0.
	    do i1=1,nbin
	      sum=sum+covmatF(i,i1)*invcovmatF(i1,j)
	    end do
	    check(i,j)=sum
	  end do
	  if (i.eq.j) then
	     if (abs(check(i,j)-1.).gt.1.d-4) 
     &		write(*,*) 'i,j=', i, j, check(i,j)
     	  else
	     if (abs(check(i,j)).gt.1.d-4) 
     &		write(*,*) 'i,j=', i, j, check(i,j)     	  
	  end if
	  	  
	end do	

	chisq=0.
	 	  
	do i=1,nbin
	  do j=1,nbin
	    chisq=chisq+(ysn(i)-y0a(i))*invcovmatF(i,j)*
     &			(ysn(j)-y0a(j))
	  end do
	end do
	
	write(*,*) 'chisq_SNe=',chisq
	
101	format(i5, a9,8e11.3)	

	stop
	end
			

!cccccccccccccccccccccccccccccccccccccccccccccccccccccc

      subroutine matinvert(mat,ident,indx,n,np)
      implicit none
      integer  n,np,indx(np)
      real*8  mat(np,np),ident(np,np)
 
      integer  i,j
      real*8   d,norm
      external ludcmp, lubksb
 
!c ... set up identity matrix
 
      do i=1,n
        do j=1,n
          ident(i,j)=0.
        end do
        ident(i,i)=1.
      end do
 
!c ... divide by mean eigenvalue of the matrix to stabilise the inversion
 
      norm=0.
      do i=1,n
        norm=norm+mat(i,i)
      end do
      norm=norm/real(n)
 
      do i=1,n
        do j=1,n
          mat(i,j)=mat(i,j)/norm
        end do
      end do
 
!c ... perform LU decomposition and backsubtitution to invert matrix
 
      call ludcmp(mat,n,np,indx,d)
      do j=1,n
        call lubksb(mat,n,np,indx,ident(1,j))
      end do
 
!c ... normalise the inverse
 
      do i=1,n
        do j=1,n
          mat(i,j)=ident(i,j)/norm
        end do
      end do
       
      return
      end
 

      SUBROUTINE ludcmp(a,n,np,indx,d)
      INTEGER n,np,indx(n),NMAX
      REAL*8  d,a(np,np),TINY
      PARAMETER (NMAX=1000,TINY=1.0d-20)
      INTEGER i,imax,j,k
      REAL*8   aamax,dum,sum,vv(NMAX)
      
      singularmat=0
      
      d=1.
      do 12 i=1,n
        aamax=0.d0
        do 11 j=1,n
          if (abs(a(i,j)).gt.aamax) aamax=abs(a(i,j))
11      continue
        if (aamax.eq.0.) pause 'singular matrix in ludcmp'
        
        vv(i)=1.0/aamax
12    continue
      do 19 j=1,n
        do 14 i=1,j-1
          sum=a(i,j)
          do 13 k=1,i-1
            sum=sum-a(i,k)*a(k,j)
13        continue
          a(i,j)=sum
14      continue
        aamax=0.
        do 16 i=j,n
          sum=a(i,j)
          do 15 k=1,j-1
            sum=sum-a(i,k)*a(k,j)
15        continue
          a(i,j)=sum
          dum=vv(i)*abs(sum)
          if (dum.ge.aamax) then
            imax=i
            aamax=dum
          endif
16      continue
        if (j.ne.imax)then
          do 17 k=1,n
            dum=a(imax,k)
            a(imax,k)=a(j,k)
            a(j,k)=dum
17        continue
          d=-d
          vv(imax)=vv(j)
        endif
        indx(j)=imax
        if(a(j,j).eq.0.0)a(j,j)=TINY
        if(j.ne.n)then
          dum=1.0/a(j,j)
          do 18 i=j+1,n
            a(i,j)=a(i,j)*dum
18        continue
        endif
19    continue
  
20      return
      END

!c-----*-----------------------------------------------------------------

      SUBROUTINE lubksb(a,n,np,indx,b)
      INTEGER n,np,indx(n)
      REAL*8  a(np,np),b(n)
      INTEGER i,ii,j,ll
      REAL*8  sum
      ii=0
      do 12 i=1,n
        ll=indx(i)
        sum=b(ll)
        b(ll)=b(i)
        if (ii.ne.0)then
          do 11 j=ii,i-1
            sum=sum-a(i,j)*b(j)
11        continue
        else if (sum.ne.0.0) then
          ii=i
        endif
        b(i)=sum
12    continue
      do 14 i=n,1,-1
        sum=b(i)
        do 13 j=i+1,n
          sum=sum-a(i,j)*b(j)
13      continue
        b(i)=sum/a(i,i)
14    continue
      return
      END

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
       
	subroutine intg (a,b,f,eps,ans,del)
	implicit real  (a-h,o-z)
	integer nmin
	common /nmin/ nmin
	external f
	
	dimension t(20,20)	!	matrix of approximations
	c = 0.5d0 * (b + a)	!	center
	d = 0.5d0 * (b - a)	!	width
	n = 2			!	level
	m = 1			!	counts something or other
	e = 1.0		!	something like a step size
	t(1,1) = 0.0		!	(cf. HP comment above)
	t(1,2) = 2.0 * d * f(c)
	t(2,1) = 0.75d0 * t(1,2)
	
1000	continue
	n = n + 1
	m = m * 2
	e = e * 0.5d0
	s = 0.0
	do 100 j = 2, m, 2
	y = dble(j-1) * e
	x = 0.5d0*y*(3.0 - y**2)
	s = s + (1.0 - y**2) * (f(c-d*x) + f(c+d*x))
100	continue
	t(n,1) = 1.5d0*s*d*e + 0.5d0*t(n-1,1)

	p = 1.0
	do 200 k = 1, n-1
	p = p * 4.0
	i = n + 1 - k
	t(i-1,k+1) = t(i,k) + (t(i,k) - t(i-1,k))/(p-1.0)
200	continue

	ans = t(1,n)
	del = abs(t(1,n)-t(2,n-1))
	if (n .lt. nmin) go to 1000
	if (n .ge. 20)	go to 900
	if (abs(del) .le. eps*abs(ans))  go to 900
	go to 1000

900	continue

	return
	end

!------------------------------------------------------------------

	function invEz(z)
	implicit none
	real Omegam, Omegak, z, w, w0, wa
	real sum0, invEz, OmegaX, Xz, ex
	common /cospar/ Omegam, Omegak, w0, wa
	external ex
							
	w=1.+z

	OmegaX=1.-Omegam-Omegak
	Xz= ex(3.*(1.+w0+wa)*log(w)+ 3.*wa*(1./w-1.))	!w0+wa(1-a)
	sum0=Omegam*w**3+ Omegak*w**2+ OmegaX*Xz

	invEz = 1.0/sqrt(sum0)
	
	return
	end 
		
!------------------------------------------------------------------

	function ex(x)
	implicit none
	real x, ex
	
	if (x.lt.-15.) then
	  ex=0.
	else
	  ex=exp(x)
	end if      

	return
	end


